(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) : typeof define === 'function' && define.amd ? define(['exports'], factory) : (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.AsyncComputed = {}));
})(this, function (exports) {
  'use strict';

  const prefix = '_async_computed$';

  /** */
  function setAsyncState(vm, stateObject, state) {
    vm.$set(vm.$data._asyncComputed[stateObject], 'state', state);
    vm.$set(vm.$data._asyncComputed[stateObject], 'updating', state === 'updating');
    vm.$set(vm.$data._asyncComputed[stateObject], 'error', state === 'error');
    vm.$set(vm.$data._asyncComputed[stateObject], 'success', state === 'success');
  }

  /** */
  function getAsyncState(vm, stateObjectKey) {
    return vm.$data._asyncComputed[stateObjectKey].state;
  }

  /** */
  function getterOnly(fn) {
    if (typeof fn === 'function') return fn;

    return fn.get;
  }

  /** */
  function hasOwnProperty(object, property) {
    return Object.prototype.hasOwnProperty.call(object, property);
  }

  /** */
  function validOpt({ data, asyncComputed, computed }) {
    var dataKeys = [];
    const computedKeys = Object.keys(computed || {});

    if (typeof data !== 'function' && data instanceof Object) dataKeys = Object.keys(data || {});

    for (const key in asyncComputed) {
      if (dataKeys.includes(key)) warn('data', key);else if (computedKeys.includes(key)) warn('computed', key);
    }
  }

  /**
   * 对冲突的键发起警告
   * @param {String} type 警告的类型
   * @param {String} key 键名称
   */
  function warn(type, key) {
    console.warn(`[Vue warn] asyncComputed 属性名称冲突: ${type} 中已定义属性 ${key}`);
  }

  /**
   * 检查计算属性的状态是否为全为指定状态
   * @param {String|Array} props 指定的属性名称，可以为属性名，或者属性名的数组
   * @param {Enum<String>} status 指定的状态，枚举类型，要求是 'updating', 'success', 'error'，三者之一
   * @return {Boolean} 是否符合指定状态，是为true，否则为false
   */
  function checkAscPropsStatus(props, status) {
    var innerProps;

    if (typeof props === 'string') {
      innerProps = [props];
    } else if (Array.isArray(props)) {
      innerProps = props;
    } else throw new Error('要求传入正确的属性名称！属性名称应为字符串，或者包含的字符串的数组。');

    // 收集对应属性的变化
    innerProps.forEach(prop => {
      this[prop], this[prefix + prop];
    });

    return innerProps.every(prop => status === getAsyncState(this, prop));
  }

  const DidNotUpdate = typeof Symbol === 'function' ? Symbol('did-not-update') : {};

  const getGetterWithShouldUpdate = (asyncProprety, currentGetter) => {
    return function getter() {
      return asyncProprety.shouldUpdate.call(this) ? currentGetter.call(this) : DidNotUpdate;
    };
  };

  const shouldNotUpdate = value => DidNotUpdate === value;

  /** */
  function isComputedLazy(item) {
    return hasOwnProperty(item, 'lazy') && item.lazy;
  }

  /** */
  function isLazyActive(vm, key) {
    return vm[lazyActivePrefix + key];
  }

  const lazyActivePrefix = 'async_computed$lazy_active$',
        lazyDataPrefix = 'async_computed$lazy_data$';
  /** */
  function initLazy(data, key, value) {
    data[lazyActivePrefix + key] = false;
    data[lazyDataPrefix + key] = value;
  }

  /** */
  function makeLazyComputed(key) {
    return {
      get() {
        this[lazyActivePrefix + key] = true;

        return this[lazyDataPrefix + key];
      },
      set(value) {
        this[lazyDataPrefix + key] = value;
      }
    };
  }

  /** */
  function silentSetLazy(vm, key, value) {
    vm[lazyDataPrefix + key] = value;
  }

  /** */
  function silentGetLazy(vm, key) {
    return vm[lazyDataPrefix + key];
  }

  /** */
  function errorHandler({ vm, key, pluginOptions, err }) {
    setAsyncState(vm, key, 'error');
    vm.$set(vm.$data._asyncComputed[key], 'exception', err);
    if (pluginOptions.errorHandler === false) return;

    const handler = pluginOptions.errorHandler === undefined ? () => {
      err.message = 'Error evaluating async computed property:' + err.message;
      throw err;
    } : pluginOptions.errorHandler;

    if (pluginOptions.useRawError) {
      return handler(err, vm, err.stack);
    } else {
      return handler(err.stack);
    }
  }

  const defaultTimeout = 6000;

  /** */
  function generateDefault(fn, pluginOptions, prefix) {
    let defaultValue = '_$undefined';

    if ('default' in fn) {
      defaultValue = fn.default;
    } else if ('default' in pluginOptions) {
      defaultValue = pluginOptions.default;
    } else if (defaultValue === '_$undefined') defaultValue = getDefaultPromise.call(this, { prefix, fn, pluginOptions });

    if (typeof defaultValue === 'function') {
      this.$once('hook:created', () => {
        const value = defaultValue.call(this, fn, pluginOptions),
              { key } = fn;

        if (isComputedLazy(fn)) {
          silentSetLazy(this, key, value);
        } else {
          this[key] = value;
        }
      });
    } else {
      return defaultValue;
    }
  }

  /** */
  function getDefaultPromise({ prefix, fn, pluginOptions }) {
    const { key, timeout = pluginOptions.timeout || defaultTimeout } = fn;

    return new Promise((resolve, reject) => {
      const timer = setTimeout(() => {
        reject(`asyncComputed 属性 “${key}” 异步执行栈超时${timeout / 1000}秒！`);
      }, timeout);

      /** */
      function handleError(err, vm) {
        try {
          resolve(errorHandler({ vm, key, pluginOptions, err }));
        } catch (newErr) {
          reject(newErr);
        }
      }

      this.$once('hook:created', async () => {
        if (isComputedLazy(fn)) {
          const unWatch = this.$watch(`${lazyActivePrefix}${key}`, async newVal => {
            if (newVal) {
              try {
                resolve((await this[`${prefix}${key}`]));
              } catch (e) {
                handleError(e, this);
              } finally {
                clearTimeout(timer);
                unWatch();
              }
            }
          });
        } else {
          try {
            resolve((await this[`${prefix}${key}`]));
          } catch (e) {
            handleError(e, this);
          } finally {
            clearTimeout(timer);
          }
        }
      });
    });
  }

  const getGetterWatchedByArray = computedAsyncProperty => function getter() {
    computedAsyncProperty.watch.forEach(key => {
      // Check if nested key is watched.
      const splittedByDot = key.split('.');

      if (splittedByDot.length === 1) {
        // If not, just access it.
        // eslint-disable-next-line no-unused-expressions
        this[key];
      } else {
        // Access the nested propety.
        try {
          let start = this;

          splittedByDot.forEach(part => {
            start = start[part];
          });
        } catch (error) {
          console.error('AsyncComputed: bad path: ', key);
          throw error;
        }
      }
    });

    return computedAsyncProperty.get.call(this);
  },
        getGetterWatchedByFunction = computedAsyncProperty => function getter() {
    computedAsyncProperty.watch.call(this);

    return computedAsyncProperty.get.call(this);
  };

  function getWatchedGetter(computedAsyncProperty) {
    if (typeof computedAsyncProperty.watch === 'function') {
      return getGetterWatchedByFunction(computedAsyncProperty);
    } else if (Array.isArray(computedAsyncProperty.watch)) {
      computedAsyncProperty.watch.forEach(key => {
        if (typeof key !== 'string') {
          throw new Error('AsyncComputed: watch elemnts must be strings');
        }
      });

      return getGetterWatchedByArray(computedAsyncProperty);
    } else {
      throw Error('AsyncComputed: watch should be function or an array');
    }
  }

  const prefix$1 = '_async_computed$',
        AsyncComputed = {
    install(Vue, pluginOptions = {}) {
      Vue.config.optionMergeStrategies.asyncComputed = Vue.config.optionMergeStrategies.computed;

      Vue.mixin(getAsyncComputedMixin(pluginOptions));

      const { checkAscPropsStatusPrefix = '$' } = pluginOptions;

      Vue.prototype[`${checkAscPropsStatusPrefix}checkAscPropsStatus`] = checkAscPropsStatus;
    }
  };

  /** */
  function getAsyncComputedMixin(pluginOptions) {
    return {
      data() {
        return {
          _asyncComputed: {}
        };
      },
      computed: {
        $asyncComputed() {
          return this.$data._asyncComputed;
        }
      },
      beforeCreate() {
        const asyncComputed = this.$options.asyncComputed || {};

        if (!Object.keys(asyncComputed).length) return;

        for (const key in asyncComputed) {
          const getter = getterFn(key, asyncComputed[key]);

          this.$options.computed[prefix$1 + key] = getter;
        }

        this.$options.data = initDataWithAsyncComputed(this.$options, pluginOptions);
      },
      created() {
        for (const key in this.$options.asyncComputed || {}) {
          const item = this.$options.asyncComputed[key];

          if (/^(_|$)/.test(key)) {
            if (isComputedLazy(item)) {
              silentSetLazy(this, key, silentGetLazy(this, key));
            } else {
              this[key] = this.$data[key];
            }
          }

          handleAsyncComputedPropetyChanges(this, key, pluginOptions);
        }
      }
    };
  }

  const AsyncComputedMixin = getAsyncComputedMixin();

  /** */
  function handleAsyncComputedPropetyChanges(vm, key, pluginOptions) {
    let promiseId = 0;
    const watcher = newPromise => {
      const thisPromise = ++promiseId;

      if (shouldNotUpdate(newPromise)) return;

      if (!newPromise || !newPromise.then) {
        newPromise = Promise.resolve(newPromise);
      }

      if (['success', 'error'].includes(vm.$data._asyncComputed[key].state)) vm[key] = newPromise;
      // 此处添加 vm[key] = newPromise 效果与设置 default 一致，但会导致 watch 异步计算属性时首次 watch 会被执行两次（即触发两次 update）
      // if (!(pluginOptions.hasOwnProperty('default') || vm.$options.asyncComputed[key].hasOwnProperty('default'))) vm[key] = newPromise

      setAsyncState(vm, key, 'updating');

      newPromise.then(value => {
        if (thisPromise !== promiseId) return;
        setAsyncState(vm, key, 'success');
        vm[key] = value;
      }).catch(err => {
        if (thisPromise !== promiseId) return;

        if (vm.$options.asyncComputed[key].hasOwnProperty('default')) {
          try {
            errorHandler({ vm, key, pluginOptions, err });
          } catch (e) {
            console.error(e);
          }
        }
      });
    };

    vm.$set(vm.$data._asyncComputed, key, {
      exception: null,
      update: () => {
        if (!vm._isDestroyed) {
          watcher(getterOnly(vm.$options.asyncComputed[key]).apply(vm));
        }
      }
    });
    setAsyncState(vm, key, 'updating');
    vm.$watch(prefix$1 + key, watcher, { immediate: true });
  }

  /** */
  function initDataWithAsyncComputed(options, pluginOptions) {
    const { data: optionData, computed } = options,
          asyncComputed = options.asyncComputed || {};

    return function vueAsyncComputedInjectedDataFn(vm) {
      const data = (typeof optionData === 'function' ? optionData.call(this, vm) : optionData) || {};

      validOpt({ computed, data, asyncComputed });

      for (const key in asyncComputed) {
        const item = this.$options.asyncComputed[key];

        item.key = key;

        const value = generateDefault.call(this, item, pluginOptions, prefix$1);

        if (isComputedLazy(item)) {
          initLazy.call(this, data, key, value);
          this.$options.computed[key] = makeLazyComputed(key);
        } else {
          data[key] = value;
        }
      }

      return data;
    };
  }

  /** */
  function getterFn(key, fn) {
    if (typeof fn === 'function') return fn;

    let getter = fn.get;

    if (hasOwnProperty(fn, 'watch')) {
      getter = getWatchedGetter(fn);
    }

    if (hasOwnProperty(fn, 'shouldUpdate')) {
      getter = getGetterWithShouldUpdate(fn, getter);
    }

    if (isComputedLazy(fn)) {
      const nonLazy = getter;

      getter = function lazyGetter() {
        if (isLazyActive(this, key)) {
          return nonLazy.call(this);
        } else {
          return silentGetLazy(this, key);
        }
      };
    }

    return getter;
  }

  /* istanbul ignore if */
  if (typeof window !== 'undefined' && window.Vue) {
    // Auto install in dist mode
    window.Vue.use(AsyncComputed);
  }

  exports.AsyncComputedMixin = AsyncComputedMixin;
  exports.AsyncComputedPlugin = AsyncComputed;
  exports.default = AsyncComputed;

  Object.defineProperty(exports, '__esModule', { value: true });
});
